# A TDB file consists of a sequence of blocks, the format of each
# block is simply:
# 1 Byte: Block ID
# 2 Bytes: Block length (l):
# l Bytes: Data

import StringIO
import types
from util import *

logger = None

class Block(object):
    def __init__(self, f):
        self.bid = get_byte(f)
        self.length = get_ushort(f)
        self.data = StringIO.StringIO(f.read(self.length))
        if len(self.data.buf) < self.length:
            raise EndOfFile("End of file reached early on '%s'." % f.name)

class TDBFile(object):
    """This represents the TDBFile with its known blocks, such as
       the header, copyright block, overview block and detailed maps."""
    def __init__(self, fn):
        if type(fn) == types.FileType:
            self.f = fn
        else:
            self.f = open(filename, 'rb')
        self.header_block = None
        self.copyright_block = None
        self.overview_block = None
        self.trademark_block = None
        self.detail_blocks = []
        self.region_blocks = []
        self.unknown_block = None
        self._parse()

    def _parse(self):
        """Dispatches subparsers for *known* blocks"""
        block_parsers = {
            0x50: self.parse_header,
            0x44: self.parse_copyright,
            0x42: self.parse_overview,
            0x4C: self.parse_detail,
            0x52: self.parse_trademark,
            0x53: self.parse_regions,
            0x54: self.parse_unknown,
        }

        while self.f.read(1):
            self.f.seek(-1, 1)
            b = Block(self.f)
            if b.bid in block_parsers:
                block_parsers[b.bid](b)
            else:
                write('Unknown Block: %02X, length: %d, %s' % (b.bid, b.length, hexify(b.data.buf)))
        self.f.close()

    def parse_header(self, block):
        hb = {}
        hb['Product ID'] = get_ushort(block.data)
        hb['Family ID']  = get_ushort(block.data)

        tdb_version = get_ushort(block.data)
        hb['TDB Major Version'] = tdb_version / 100
        hb['TDB Minor Version'] = tdb_version % 100
        hb['TDB Version'] = "%d.%02d" % (hb['TDB Major Version'], hb['TDB Minor Version'])

        hb['Map Series'] = get_str(block.data)

        prod_version = get_ushort(block.data)
        hb['Product Major Version'] = prod_version / 100
        hb['Product Minor Version'] = prod_version % 100
        hb['Product Version'] = "%d.%02d" % (hb['Product Major Version'], hb['Product Minor Version'])

        hb['Map Family'] = get_str(block.data)

        block_lengths = [30, 8, 11, 9, 2, 1]
        for (e, bl) in enumerate(block_lengths):
            if block.data.tell() < len(block.data.buf):
                hb['Remainder-%d' % (e + 1)] = block.data.read(bl)

        self.header_block = hb

    def parse_copyright(self, block):
        cl = []
        while block.data.pos < len(block.data.buf):
            copyright_code = get_byte(block.data)
            where_code = get_byte(block.data)
            extra = get_short(block.data)
            copyright_string = get_str(block.data)
            if copyright_code == 0x00:
                cl.append({'Source': copyright_string.decode('latin1').encode('utf-8')})
            elif copyright_code == 0x06:
                cl.append({'Copyright': copyright_string.decode('latin1').encode('utf-8')})
            elif copyright_code == 0x07:
                cl.append({'Bitmap': copyright_string, 'Scale factor': extra})
            else:
                cl.append({'Unknown (%02X)' % copyright_code: copyright_string, 'Extra': extra})
        self.copyright_block = cl

    def parse_overview(self, block):
        ob = {}
        ob['Map Number'] = get_uint(block.data)
        ob['Parent Map'] = get_uint(block.data)
        ob['Latitude North'] = "%7.4f" % todegrees(get_int(block.data))
        ob['Longitude East'] = "%7.4f" % todegrees(get_int(block.data))
        ob['Latitude South'] = "%7.4f" % todegrees(get_int(block.data))
        ob['Longitude West'] = "%7.4f" % todegrees(get_int(block.data))
        ob['Description'] = get_str(block.data)
        if block.data.tell() < len(block.data.buf):
            ob['Remainder'] = block.data.read()
        self.overview_block = ob

    def parse_detail(self, block):
        db = {}
        db['Map Number'] = get_uint(block.data)
        db['Parent Map'] = get_uint(block.data)
        db['Latitude North'] = "%7.4f" % todegrees(get_int(block.data))
        db['Longitude East'] = "%7.4f" % todegrees(get_int(block.data))
        db['Latitude South'] = "%7.4f" % todegrees(get_int(block.data))
        db['Longitude West'] = "%7.4f" % todegrees(get_int(block.data))
        db['Description'] = get_str(block.data).decode('latin1').encode('utf-8')
        # Unknown
        db['Unknown-1'] = get_short(block.data)
        db['Unknown-2'] = get_short(block.data)
        # Could it be that these values have changed in v4?
        db['RGN Size'] = get_uint(block.data)
        db['TRE Size'] = get_uint(block.data)
        db['LBL Size'] = get_uint(block.data)
        db['Unknown-3'] = get_byte(block.data)
        if block.data.tell() < len(block.data.buf):
            db['Remainder'] = block.data.read()
        self.detail_blocks.append(db)

    def parse_regions(self, block):
        if len(block.data.buf) < 4:
            return
        while block.data.tell() < len(block.data.buf):
            get_byte(block.data)
            number = get_byte(block.data)
            s = get_str(block.data)
            self.region_blocks.append((number, s))

    def parse_unknown(self, block):
        s = block.data.read()
        self.unknown_block = {'Hex' : hexify(s)}

    def parse_trademark(self, block):
        tb = {}
        block.data.seek(1, 1)
        tb['Trademark'] = get_str(block.data)
        if block.data.tell() < len(block.data.buf):
            tb['Remainder'] = block.data.read()
        self.trademark_block = tb

    def print_header(self):
        for f in ['TDB Version', 'Product ID', 'Family ID', 'Map Series', 'Map Family', 'Product Version']:
            write("%-20s%s" % (f + ':', str(self.header_block[f])))
        for n in range(1, 7):
            b = 'Remainder-%d' % n
            if (b)  in self.header_block:
                write("%-20s%s" % (b + ' (' + str(len(self.header_block[b])) + "b):",  hexify(self.header_block[b])))
    def print_copyright(self):
        for c in self.copyright_block:
            write('\n        '.join(["%-20s%s" % (f + ":", str(c[f])) for f in c]))

    def print_trademark(self):
        write("%-20s%s" % ('Trademark:', self.trademark_block['Trademark'].encode('utf-8')))
        if 'Remainder' in self.trademark_block:
            write("%-20s%s" % ('Remainder (' + str(len(self.trademark_block['Remainder'])) + "b):",  hexify(self.trademark_block['Remainder'])))

    def print_overview(self):
        logger.info('Overview map:')
        for f in ['Map Number', 'Parent Map', 'Latitude North', 'Longitude East',
            'Latitude South', 'Longitude West', 'Description']:
            logger.info("    %-20s%s" % (f + ':', str(self.overview_block[f])))
        if 'Remainder' in self.overview_block:
            write("%-20s%s" % ('Remainder (' + str(len(self.overview_block['Remainder'])) + "b):",  hexify(self.overview_block['Remainder'])))

    def print_detail_blocks(self):
        logger.debug('Detailed maps:')
        for db in self.detail_blocks:
            logger.debug("   %-20s%s" % ('Map Number:', str(db['Map Number'])))
            logger.debug("   %-20s%s" % ('Parent Map:', str(db['Parent Map'])))
            logger.debug("   %-20s%s" % ('Description:', str(db['Description'])))
            logger.debug("   N: %s, S: %s, W: %s, E: %s" % (db['Latitude North'], db['Latitude South'],
                 db['Longitude West'], db['Longitude East']))
            logger.debug("   RGN: %d, TRE: %d, LBL: %d" % (db['RGN Size'], db['TRE Size'], db['LBL Size']))
            logger.debug("")
        if 'Remainder' in db:
            write("%-20s%s" % ('Remainder (' + str(len(db['Remainder'])) + "b):", hexify(db['Remainder'])))


    def print_regions(self):
        write('Regions covered:')
        for r in self.region_blocks:
            write("    %02d: %s" % r)

    def print_unknown_block(self):
        logger.info('Unknown block:')
        for f in ['Hex']:
            logger.info("    %-20s%s" % (f + ':', str(self.unknown_block[f])))

    def print_dump(self):
        """Gives a short textual description of the data."""
        if self.header_block:
            self.print_header()
        else:
            logger.warning('TDB file contains no header block.')
        write("")
        if self.copyright_block:
            self.print_copyright()
        else:
            logger.info('TDB file contains no copyright block.')
        write("")
        if self.trademark_block:
            self.print_trademark()
            write("")
        else:
            logger.info('TDB file contains no trademark block.')
            logger.info("")
        if self.region_blocks:
            self.print_regions()
            write("")
        else:
            logger.info('TDB file contains no region information block.')
            write("")
        if self.overview_block:
            self.print_overview()
        else:
            logger.warning('TDB file contains no overview map.')
        logger.info("")
        if self.detail_blocks:
            self.print_detail_blocks()
        else:
            logger.warning('TDB file contains no detail blocks.')

        logger.info("")
        if self.unknown_block:
            self.print_unknown_block()

def setLogger(l):
    global logger
    logger = l
